"""

USAGE:
    Format:
        format1
            clpaws eipctl --associate -n EC2_NETWORK_IF_ID -a ALLOCATION_ID
        format2
            clpaws eipctl --disassociate -a ALLOCATION_ID
        format3
            clpaws eipctl --check-eip -n EC2_NETWORK_IF_ID -a ALLOCATION_ID

    Argument:
        --associate          : associate EIP
        --disassociate       : disassociate EIP
        --check-eip          : check EIP
        --debug              : output debug log (optional)
        -n EC2_NETWORK_IF_ID : EC2 Network Interface ID
        -a ALLOCATION_ID     : Allocation ID
        -t TIMEOUT           : timeout (optional)

    ReturnCode:
        value of clpaws.errors.ERRORNO

    Example:
        example1
            clpaws eipctl --associate -n eni-00000000 -a eipalloc-00000000
        example2
            clpaws eipctl --disassociate -a eipalloc-00000000
        example3
            clpaws eipctl --check-eip -n eni-00000000 -a eipalloc-00000000
"""


import sys
import optparse
import clpaws.utils as util
import clpaws.errors as error
import logging

# global varivable
loglevel = logging.WARNING

def _check_option():
    exclusive_options = ('associate', 'disassociate', 'check_eip')

    parser = optparse.OptionParser()

    # exclusive long options
    parser.add_option('--associate',
                        dest='associate',
                        action='store_true',
                        help='associate EIP')
    parser.add_option('--disassociate',
                        dest='disassociate',
                        action='store_true',
                        help='disassociate EIP')
    parser.add_option('--check-eip',
                        dest='check_eip',
                        action='store_true',
                        help='check EIP')

    # optional long option
    parser.add_option('--debug',
                        dest='debug',
                        action='store_true',
                        help='output debug log')

    # required short options
    parser.add_option('-n',
                        dest='ec2_network_if_id',
                        help='EC2 Network Interface ID',
                        metavar='EC2_NETWORK_IF_ID')
    parser.add_option('-a',
                        dest='allocation_id',
                        help='Allocation ID',
                        metavar='ALLOCATION_ID')

    # optional short option
    parser.add_option('-t',
                        dest='timeout',
                        help='timeout',
                        metavar='TIMEOUT')

    (options, args) = parser.parse_args()

    toggled_options = [name for name in exclusive_options
                        if getattr(options, name)]

    if not len(toggled_options) == 1:
        parser.error('only one argument is required '
                     'among arguments "--associate", "--disassociate", '
                     'and "--check-eip".')

    if not options.allocation_id:
        parser.error('argument "-a" is required.')

    if options.associate and not (options.allocation_id and
                                    options.ec2_network_if_id):
        parser.error('2 arguments "-a" and "-n" are required '
                     'when argument "--associate" is specified.')

    if options.disassociate and not options.allocation_id:
        parser.error('argument "-a" is required '
                        'when argument "--disassociate" is specified.')

    if options.check_eip and not (options.allocation_id and
                                    options.ec2_network_if_id):
        parser.error('2 arguments "-a" and "-n" are required '
                     'when argument "--check-eip" is specified.')

    return options


def associate_eip(ec2_network_if_id, allocation_id, ec2_cmdopt, timeout):
    logging.debug('Start associate_eip.')

    addresses = _describe_addresses(allocation_id, ec2_cmdopt, timeout)
    if (addresses and len(addresses) == 1):
        # target eip is found
        address = addresses[0]
        if address.get('NetworkInterfaceId') == ec2_network_if_id:
            # target eip is already associated; end
            return

    _associate_address(ec2_network_if_id, allocation_id, ec2_cmdopt, timeout)

    logging.debug('End associate_eip.')


def disassociate_eip(allocation_id, ec2_cmdopt, timeout):
    logging.debug('Start disassociate_eip.')

    addresses = _describe_addresses(allocation_id, ec2_cmdopt, timeout)
    if (addresses and len(addresses) == 1):
        association = addresses[0].get('AssociationId')
        if association:
            _disassociate_address(association, ec2_cmdopt, timeout)

    logging.debug('End disassociate_eip.')


def check_eip(ec2_network_if_id, allocation_id, ec2_cmdopt, timeout):
    logging.debug('Start check_eip.')

    addresses = _describe_addresses(allocation_id, ec2_cmdopt, timeout)
    if (addresses and len(addresses) == 1):
        # target eip is found
        address = addresses[0]
        if not (address.get('NetworkInterfaceId') == ec2_network_if_id):
            err_msg = 'EIP {0} does not exist on ENI {1}'.format(allocation_id,
                                                            ec2_network_if_id)
            raise error.AWSCLIError(error.ERRORNO['check_eip_error'],
                                err_msg)
    else:
        err_msg = 'EIP {0} does not exist'.format(allocation_id)
        raise error.AWSCLIError(error.ERRORNO['check_eip_error'],
                                err_msg)

    logging.debug('End check_eip.')


def _describe_addresses(allocation_id, ec2_cmdopt, timeout):
    if not ec2_cmdopt:
        cmdline = (util.CLPAWS_CMD + ' ec2 describe-addresses --allocation-ids "{0}" '
                   '--query \'Addresses[].{{NetworkInterfaceId: NetworkInterfaceId, AssociationId: AssociationId}}\''
                   .format(allocation_id))
    else:
        cmdline = (util.CLPAWS_CMD + ' ec2 describe-addresses --allocation-ids "{0}" '
                   '--query \'Addresses[].{{NetworkInterfaceId: NetworkInterfaceId, AssociationId: AssociationId}}\' {1}'
                   .format(allocation_id, ec2_cmdopt))
    result = util.execute_awscli(cmdline, timeout)
    return result


def _associate_address(ec2_network_if_id, allocation_id, ec2_cmdopt, timeout):
    if not ec2_cmdopt:
        cmdline = (util.CLPAWS_CMD + ' ec2 associate-address --allow-reassociation --network-interface-id "{0}" '
                   '--allocation-id "{1}"'
                   .format(ec2_network_if_id, allocation_id))
    else:
        cmdline = (util.CLPAWS_CMD + ' ec2 associate-address --allow-reassociation --network-interface-id "{0}" '
                   '--allocation-id "{1}" {2}'
                   .format(ec2_network_if_id, allocation_id, ec2_cmdopt))
    util.execute_awscli(cmdline, timeout, err_msg='EIP failed to associate.')


def _disassociate_address(association, ec2_cmdopt, timeout):
    if not ec2_cmdopt:
        cmdline = (util.CLPAWS_CMD + ' ec2 disassociate-address --association-id "{0}"'
                   .format(association))
    else:
        cmdline = (util.CLPAWS_CMD + ' ec2 disassociate-address --association-id "{0}" {1}'
                   .format(association, ec2_cmdopt))
    util.execute_awscli(cmdline, timeout, err_msg='EIP failed to disassociate.')


def main():
    global loglevel
    try:
        options = _check_option()

        if options.debug:
            loglevel = logging.DEBUG

        logging.basicConfig(level=loglevel,
                            format='%(levelname)s - %(message)s',
                            stream=sys.stdout)

        logging.debug('Start clpaws eipctl.')

        util.set_environ()

        ec2_cmdopt = util.get_awscli_cmdopt('ec2')

        if options.associate:
            associate_eip(options.ec2_network_if_id, options.allocation_id, ec2_cmdopt,
                            options.timeout)
        elif options.disassociate:
            disassociate_eip(options.allocation_id, ec2_cmdopt, options.timeout)
        elif options.check_eip:
            check_eip(options.ec2_network_if_id, options.allocation_id, ec2_cmdopt,
                        options.timeout)
        else:
            err_msg = 'Invalid option.'
            raise error.AWSCLIError(error.ERRORNO['internal_error'], err_msg)

        logging.debug('End clpaws eipctl.')

        return error.ERRORNO['success']

    except error.AWSCLIError as ex:
        util.handle_exception(ex)
        return ex.value

    except Exception as ex:
        logging.error(ex)
        return error.ERRORNO['internal_error']

if __name__ == '__main__':
    sys.exit(main())
