"""

DESCRIPTION:
    This module provides utility functions for internal use.
"""


# support both of Python2 and Python3
try:
    from ConfigParser import SafeConfigParser as ConfigParser
except ImportError:
    from configparser import ConfigParser

import re
import json
import locale
import logging
import os
import shlex
import subprocess
import threading
import clpaws.errors as error

ENCODING = locale.getpreferredencoding()

# AWS CLI command
CLPAWS_CMD='aws'

# modify according to the form of AWS CLI error message
CAUSE_REGEX = re.compile(r'A.* error .* operation: (.*)$')

SECTION_REGEX = re.compile(r'\[(?P<section>[^]]+)\]')
OPTION_REGEX = re.compile(
    r'(?P<option>[^:=][^:=]*)'
    r'\s*(?P<eq>[:=])\s*'
    r'(?P<value>.*)$'
)

CREDENTIAL_KEYS = (
    'aws_access_key_id',
    'aws_secret_access_key'
)
ENVIRONMENT_VALIABLES = {
    CREDENTIAL_KEYS[0] : 'AWS_ACCESS_KEY_ID',
    CREDENTIAL_KEYS[1] : 'AWS_SECRET_ACCESS_KEY'
}

LOGGING_BYTES = 1024

SET_ENVIRON_LEGACY_FLAG = False

if os.name == 'nt': # Windows
    CONFIG_FILE = os.environ['SystemDrive'] + '\\Users\\Administrator\\.aws\\config'
    CREDENTIAL_FILE = os.environ['SystemDrive'] + '\\Users\\Administrator\\.aws\\credentials'
else: # Linux
    if 'HOME' in os.environ:
        AWS_CONFIG_PATH = os.environ['HOME'] + '/.aws/config'
        if not os.path.exists(AWS_CONFIG_PATH):
            del os.environ['HOME']


def handle_exception(ex):
    logging.error(ex.msg)
    if ex.stdouterr:
        logging.error(ex.stdouterr)


def kill_process(proc, cmdline, flag):
    try:
        if os.name == 'nt':
            proc.kill()
        else:
            proc_file = '/proc/{0}/cmdline'.format(proc.pid)
            if os.path.exists(proc_file):
                with open(proc_file, 'r') as f:
                    args = [s for s in f.read().split('\x00') if s]
                cmdline_shlex = shlex.split(cmdline)

                # pick out aws cli subcommand starting with 'ec2'
                if 'ec2' in cmdline_shlex:
                    cmdline_shlex = cmdline_shlex[cmdline_shlex.index('ec2'):]
                if 'ec2' in args:
                    args = args[args.index('ec2'):]

                if cmdline_shlex == args:
                    proc.kill()
                else:
                    pass
            else:
                pass
    finally:
        flag['timeout'] = True


def execute_command(cmdline, timeout=None):
    logging.debug('Start execute_command.')

    flag = {'timeout' : False}
    proc = subprocess.Popen(shlex.split(cmdline),
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE)
    if not timeout is None:
        timeout_sec = int(timeout)

        if timeout_sec < 0:
            err_msg = 'timeout {0} is invalid.'.format(timeout_sec)
            raise error.CommandError(error.ERRORNO['argument_error'], err_msg)

        timer = threading.Timer(timeout_sec, kill_process, [proc, cmdline, flag])
        timer.start()
        comm = proc.communicate()
        timer.cancel()
        if flag['timeout'] == True:
            err_msg = 'Command ({0}) was timeout.'.format(cmdline)
            raise error.CommandError(error.ERRORNO['timeout_error'], err_msg)
        else:
            return_value = proc.returncode
            logging.debug('End execute_command.')
            return comm, return_value
    else:
        comm = proc.communicate()
        return_value = proc.returncode
        logging.debug('End execute_command.')
        return comm, return_value


def execute_awscli(cmdline, timeout=None, err_msg=None):
    logging.debug('Start execute_awscli.')

    try:
        comm, return_value = execute_command(cmdline, timeout)
    except error.CommandError as e:
        if e.value == error.ERRORNO['timeout_error']:
            err_msg = 'AWS CLI ({0}) was timeout.'.format(cmdline)
            raise error.AWSCLIError(e.value, err_msg)
        else:
            raise error.AWSCLIError(e.value, e.msg)

    if return_value == 0:
        output = comm[0].decode(ENCODING)
        if len(output) >= LOGGING_BYTES:
            logging.debug(output[0:LOGGING_BYTES])
        else:
            logging.debug(output)
        logging.debug('End execute_awscli.')
        if len(output) > 0:
            return json.loads(output)
        else:
            return json.loads('{}')
    else:
        if comm[1] is None:
            output = ""
        else:
            output = comm[1].decode(ENCODING)
            match = CAUSE_REGEX.search(output)
            if not match is None:
                cause = "cause: " + match.group(1) + "\n"
            else:
                cause = ""
            output = output + cause
        if not err_msg:
            err_msg = 'Failed to execute AWS CLI ({0}). '\
                        'AWS CLI returned ({1}).'.format(cmdline, return_value)

        raise error.AWSCLIError(error.ERRORNO['command_error'],
                                err_msg, output)


def execute_cfget(option, xpath, policy=None, timeout=None):
    logging.debug('Start execute_cfget. (option={0}, xpath={1}, policy={2})'.format(option, xpath, policy))

    if policy is not None:
        cmdline = 'clpcfget {0} {1} -p {2}'.format(option, xpath, policy)
    else:
        cmdline = 'clpcfget {0} {1}'.format(option, xpath)
    comm, return_value = execute_command(cmdline, timeout)
    output = comm[0].decode(ENCODING).strip()

    if return_value != 0:
        logging.debug('cmd: {0} = {1}'.format(cmdline, return_value))

        if len(output) >= LOGGING_BYTES:
            logging.debug('msg: ' + output[0:LOGGING_BYTES])
        else:
            logging.debug('msg: ' + output)

        raise error.CommandError(error.ERRORNO['command_error'], output)

    logging.debug('End execute_cfget.')
    return output


def execute_cfget_enum(xpath, policy=None, timeout=None):
    return execute_cfget('-e', xpath, policy, timeout).splitlines()


def execute_cfget_value(xpath, policy=None, timeout=None):
    return execute_cfget('-g', xpath, policy, timeout)


def _find_default_section_start(contents):
    for i in range(len(contents)):
        line = contents[i]
        if line.strip().startswith(('#', ';')):
            # ignore a comment line.
            continue
        match = SECTION_REGEX.search(line)
        if (match is not None and match.group(1) == 'default'):
            return i

    return -1


def _read_default_section_credentials(contents):
    section_start_line_num = _find_default_section_start(contents)
    if section_start_line_num < 0:
        logging.error('default section does not exist.')
        return

    section_start_line_num += 1
    i = section_start_line_num
    while i < len(contents):
        line = contents[i]
        match = OPTION_REGEX.search(line)
        if match is not None:
            key_name = match.group(1).strip()
            value_name = match.group(3).strip()
            if key_name in CREDENTIAL_KEYS:
                os.environ[ENVIRONMENT_VALIABLES[key_name]] = value_name
            i += 1
        elif SECTION_REGEX.search(line) is not None:
            if line.strip().startswith(('#', ';')):
            # ignore a comment line.
                i += 1
                continue
            else:
                # We are out of default section.
                return
        else:
            i += 1

    return


def set_cerdential_environ(filename):
    logging.debug('Start set_cerdential_env.')

    if os.path.exists(filename):
        with open(filename, 'r') as f:
            contents = f.readlines()
        _read_default_section_credentials(contents)
    else:
        logging.debug('cerdential file does not exist.')

    logging.debug('End set_cerdential_env.')

def set_environ_legacy():
    logging.debug('Start set_environ_legacy.')
    set_environ_file()

    os.environ['AWS_DEFAULT_OUTPUT'] = 'json'

    if os.name == 'nt': # Windows
        os.environ['AWS_CONFIG_FILE'] = CONFIG_FILE
        set_cerdential_environ(CREDENTIAL_FILE)

    else: # Linux
        os.environ['PATH'] += ':/usr/local/bin' # for AWS CLI

    logging.debug('End set_environ_legacy.')

def set_environ():
    if SET_ENVIRON_LEGACY_FLAG:
        set_environ_legacy()
        return

    logging.debug('Start set_environ.')

    if os.name == 'nt': # Windows
        os.environ['AWS_CONFIG_FILE'] = CONFIG_FILE
        os.environ['AWS_SHARED_CREDENTIALS_FILE'] = CREDENTIAL_FILE

    else: # Linux
        os.environ['PATH'] += ':/usr/local/bin' # for AWS CLI

    if not set_environ_cfget():
        set_environ_file()

    os.environ['AWS_DEFAULT_OUTPUT'] = 'json'

    logging.debug('End set_environ.')

def set_environ_cfget():
    logging.debug('Start set_environ_cfget.')

    xpath = '/root/cluster/cloud/aws/envs/env'

    try:
        ids = execute_cfget_enum(xpath)
    except Exception as e:
        logging.debug('{0}: {1}'.format(type(e), str(e.args)))
        return False

    for id in ids:
        xpath = '/root/cluster/cloud/aws/envs/env@{0}'.format(id)

        try:
            name = execute_cfget_value(xpath + '/name')
            value = execute_cfget_value(xpath + '/value')
        except Exception as e:
            logging.debug('env{0}: {1}: {2}'.format(id, type(e), str(e.args)))
            continue

        logging.debug('env{0}: set: \'{1}\' = \'{2}\''.format(id, name, value))

        if name == "":
            logging.debug('env{0}: empty name'.format(id, name, value))
            continue

        os.environ[name] = value

    logging.debug('End set_environ_cfget.')
    return True

def set_environ_file():
    logging.debug('Start set_environ_file.')
    conffilepath = os.path.abspath(os.path.dirname(__file__))
    conffilepath = conffilepath.rstrip(os.sep)
    conffilepath = conffilepath + '_setting.conf'

    if not os.path.isfile(conffilepath) :
        logging.debug("%s is nothing." % (conffilepath))
        return

    try:
        conffile = ConfigParser()
        conffile.optionxform = str
        conffile.read(conffilepath)

        items = conffile.items('ENVIRONMENT')
        for item in items :
            info_msg = 'key=[' + item[0] + '], val=[' + item[1] + ']'
            if len(info_msg) >= LOGGING_BYTES :
                logging.info(info_msg[0:LOGGING_BYTES])
            else :
                logging.info(info_msg)
            os.environ[item[0]] = item[1]

        if 'CLP_AWS_ERR_FORMAT' in os.environ:
            global CAUSE_REGEX
            CAUSE_REGEX = re.compile(r'' + os.environ['CLP_AWS_ERR_FORMAT'] + '')
        if 'CLP_AWS_CMD' in os.environ:
            global CLPAWS_CMD
            CLPAWS_CMD = os.environ['CLP_AWS_CMD']
    except Exception as e:
        logging.debug('Failed to read clpaws_setting.conf' + str(e.args))

    logging.debug('End set_environ_file.')

def get_awscli_cmdopt(awssvc):
    logging.debug('Start get_awscli_cmdopt.')

    xpath = '/root/cluster/cloud/aws/cmdopt/{0}'.format(awssvc)
    policy = 'cluster'

    try:
        awscmdopt = execute_cfget_value(xpath, policy)
        if not awscmdopt:
            return
        if ';' in awscmdopt or '&&' in awscmdopt or '||' in awscmdopt or '`' in awscmdopt:
            logging.error('awscli_cmdopt: contains the special characters.')
            logging.error('awscli_cmdopt: svcname = {0}: cmdopt = \'{1}\''.format(awssvc, awscmdopt))
            return
        if '--output' in awscmdopt:
            logging.error('awscli_cmdopt: contains the special output option.')
            logging.error('awscli_cmdopt: svcname = {0}: cmdopt = \'{1}\''.format(awssvc, awscmdopt))
            return
    except Exception as e:
        logging.debug('{0}: {1}'.format(type(e), str(e.args)))
        return

    logging.debug('awscli_cmdopt: svcname = {0}: cmdopt = \'{1}\''.format(awssvc, awscmdopt))

    logging.debug('End get_awscli_cmdopt.')
    return awscmdopt